@@ -291,6 +291,76 @@ gather_clone_modes (ScreenInfo *info)
     info->clone_modes = (GnomeRRMode **)g_ptr_array_free (result, FALSE);
 }
 
+static gboolean
+fill_screen_info_from_resources (ScreenInfo *info,
+				 XRRScreenResources *resources,
+				 GError **error)
+{
+    int i;
+    GPtrArray *a;
+    GnomeRRCrtc **crtc;
+    GnomeRROutput **output;
+
+    info->resources = resources;
+
+    /* We create all the structures before initializing them, so
+     * that they can refer to each other.
+     */
+    a = g_ptr_array_new ();
+    for (i = 0; i < resources->ncrtc; ++i)
+    {
+	GnomeRRCrtc *crtc = crtc_new (info, resources->crtcs[i]);
+
+	g_ptr_array_add (a, crtc);
+    }
+    g_ptr_array_add (a, NULL);
+    info->crtcs = (GnomeRRCrtc **)g_ptr_array_free (a, FALSE);
+
+    a = g_ptr_array_new ();
+    for (i = 0; i < resources->noutput; ++i)
+    {
+	GnomeRROutput *output = output_new (info, resources->outputs[i]);
+
+	g_ptr_array_add (a, output);
+    }
+    g_ptr_array_add (a, NULL);
+    info->outputs = (GnomeRROutput **)g_ptr_array_free (a, FALSE);
+
+    a = g_ptr_array_new ();
+    for (i = 0;  i < resources->nmode; ++i)
+    {
+	GnomeRRMode *mode = mode_new (info, resources->modes[i].id);
+
+	g_ptr_array_add (a, mode);
+    }
+    g_ptr_array_add (a, NULL);
+    info->modes = (GnomeRRMode **)g_ptr_array_free (a, FALSE);
+
+    /* Initialize */
+    for (crtc = info->crtcs; *crtc; ++crtc)
+    {
+	if (!crtc_initialize (*crtc, resources, error))
+	    return FALSE;
+    }
+
+    for (output = info->outputs; *output; ++output)
+    {
+	if (!output_initialize (*output, resources, error))
+	    return FALSE;
+    }
+
+    for (i = 0; i < resources->nmode; ++i)
+    {
+	GnomeRRMode *mode = mode_by_id (info, resources->modes[i].id);
+
+	mode_initialize (mode, &(resources->modes[i]));
+    }
+
+    gather_clone_modes (info);
+
+    return TRUE;
+}
+
 static gboolean
 fill_out_screen_info (Display *xdisplay,
 		      Window xroot,
@@ -303,6 +373,43 @@ fill_out_screen_info (Display *xdisplay,
     g_assert (xdisplay != NULL);
     g_assert (info != NULL);
 
+    /* First update the screen resources */
+
+    if (needs_reprobe)
+        resources = XRRGetScreenResources (xdisplay, xroot);
+    else
+    {
+	/* XRRGetScreenResourcesCurrent is less expensive than
+	 * XRRGetScreenResources, however it is available only
+	 * in RandR 1.3 or higher
+	 */
+#if (RANDR_MAJOR > 1 || (RANDR_MAJOR == 1 && RANDR_MINOR >= 3))
+        /* Runtime check for RandR 1.3 or higher */
+        if (info->screen->rr_major_version == 1 && info->screen->rr_minor_version >= 3)
+            resources = XRRGetScreenResourcesCurrent (xdisplay, xroot);
+        else
+            resources = XRRGetScreenResources (xdisplay, xroot);
+#else
+        resources = XRRGetScreenResources (xdisplay, xroot);
+#endif
+    }
+
+    if (resources)
+    {
+	if (!fill_screen_info_from_resources (info, resources, error))
+	    return FALSE;
+    }
+    else
+    {
+	g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_RANDR_ERROR,
+		     _("could not get the screen resources (CRTCs, outputs, modes)"));
+	return FALSE;
+    }
+
+    /* Then update the screen size range.  We do this after XRRGetScreenResources() so that
+     * the X server will already have an updated view of the outputs.
+     */
+
     if (needs_reprobe) {
 	gboolean success;
 
@@ -324,8 +431,6 @@ fill_out_screen_info (Display *xdisplay,
 			 _("could not get the range of screen sizes"));
             return FALSE;
         }
-
-        resources = XRRGetScreenResources (xdisplay, xroot);
     }
     else
     {
@@ -334,104 +439,9 @@ fill_out_screen_info (Display *xdisplay,
 					 &(info->max_width),
 					 &(info->min_height),
 					 &(info->max_height));
-	/* XRRGetScreenResourcesCurrent is less expensive than
-	 * XRRGetScreenResources, however it is available only
-	 * in RandR 1.3 or higher
-	 */
-#if (RANDR_MAJOR > 1 || (RANDR_MAJOR == 1 && RANDR_MINOR >= 3))
-        /* Runtime check for RandR 1.3 or higher */
-        if (info->screen->rr_major_version == 1 && info->screen->rr_minor_version >= 3)
-            resources = XRRGetScreenResourcesCurrent (xdisplay, xroot);
-        else
-            resources = XRRGetScreenResources (xdisplay, xroot);
-#else
-        resources = XRRGetScreenResources (xdisplay, xroot);
-#endif
     }
 
-    if (resources)
-    {
-	int i;
-	GPtrArray *a;
-	GnomeRRCrtc **crtc;
-	GnomeRROutput **output;
-	
-#if 0
-	g_print ("Resource Timestamp: %u\n", (guint32)resources->timestamp);
-	g_print ("Resource Configuration Timestamp: %u\n", (guint32)resources->configTimestamp);
-#endif
-	
-	info->resources = resources;
-	
-	/* We create all the structures before initializing them, so
-	 * that they can refer to each other.
-	 */
-	a = g_ptr_array_new ();
-	for (i = 0; i < resources->ncrtc; ++i)
-	{
-	    GnomeRRCrtc *crtc = crtc_new (info, resources->crtcs[i]);
-	    
-	    g_ptr_array_add (a, crtc);
-	}
-	g_ptr_array_add (a, NULL);
-	info->crtcs = (GnomeRRCrtc **)g_ptr_array_free (a, FALSE);
-	
-	a = g_ptr_array_new ();
-	for (i = 0; i < resources->noutput; ++i)
-	{
-	    GnomeRROutput *output = output_new (info, resources->outputs[i]);
-	    
-	    g_ptr_array_add (a, output);
-	}
-	g_ptr_array_add (a, NULL);
-	info->outputs = (GnomeRROutput **)g_ptr_array_free (a, FALSE);
-	
-	a = g_ptr_array_new ();
-	for (i = 0;  i < resources->nmode; ++i)
-	{
-	    GnomeRRMode *mode = mode_new (info, resources->modes[i].id);
-	    
-	    g_ptr_array_add (a, mode);
-	}
-	g_ptr_array_add (a, NULL);
-	info->modes = (GnomeRRMode **)g_ptr_array_free (a, FALSE);
-	
-	/* Initialize */
-	for (crtc = info->crtcs; *crtc; ++crtc)
-	{
-	    if (!crtc_initialize (*crtc, resources, error))
-		return FALSE;
-	}
-	
-	for (output = info->outputs; *output; ++output)
-	{
-	    if (!output_initialize (*output, resources, error))
-		return FALSE;
-	}
-	
-	for (i = 0; i < resources->nmode; ++i)
-	{
-	    GnomeRRMode *mode = mode_by_id (info, resources->modes[i].id);
-	    
-	    mode_initialize (mode, &(resources->modes[i]));
-	}
-
-	gather_clone_modes (info);
-	
-	return TRUE;
-    }
-    else
-    {
-#if 0
-	g_print ("Couldn't get screen resources\n");
-#endif
-	/* Translators: This indicates an error in X.  CRTCs is "CRT
-	 * controllers".  This error should never happen; it is just here for
-	 * completeness. */
-	g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_RANDR_ERROR,
-		     _("could not get the screen resources (CRTCs, outputs, modes)"));
-	return FALSE;
-    }
+    return TRUE;
 }
 
 static ScreenInfo *
@@ -497,11 +507,58 @@ screen_on_event (GdkXEvent *xevent,
     event_num = e->type - screen->randr_event_base;
 
     if (event_num == RRScreenChangeNotify) {
-	/* If outputs are either connected or disconnected, reprobe hardware
-	 * and/or get the screen size again.
+	/* We don't reprobe the hardware; we just fetch the X server's latest
+	 * state.  The server already knows the new state of the outputs; that's
+	 * why it sent us an event!
 	 */
-        screen_update (screen, TRUE, TRUE, NULL); /* NULL-GError */
+        screen_update (screen, TRUE, FALSE, NULL); /* NULL-GError */
+#if 0
+	/* Enable this code to get a dialog showing the RANDR timestamps, for debugging purposes */
+	{
+	    GtkWidget *dialog;
+	    XRRScreenChangeNotifyEvent *rr_event;
+	    static int dialog_num;
+
+	    rr_event = (XRRScreenChangeNotifyEvent *) e;
+
+	    dialog = gtk_message_dialog_new (NULL,
+					     0,
+					     GTK_MESSAGE_INFO,
+					     GTK_BUTTONS_CLOSE,
+					     "RRScreenChangeNotify timestamps (%d):\n"
+					     "event change: %u\n"
+					     "event config: %u\n"
+					     "event serial: %lu\n"
+					     "----------------------"
+					     "screen change: %u\n"
+					     "screen config: %u\n",
+					     dialog_num++,
+					     (guint32) rr_event->timestamp,
+					     (guint32) rr_event->config_timestamp,
+					     rr_event->serial,
+					     (guint32) screen->info->resources->timestamp,
+					     (guint32) screen->info->resources->configTimestamp);
+	    g_signal_connect (dialog, "response",
+			      G_CALLBACK (gtk_widget_destroy), NULL);
+	    gtk_widget_show (dialog);
+	}
+#endif
     }
+#if 0
+    /* WHY THIS CODE IS DISABLED:
+     *
+     * Note that in gnome_rr_screen_new(), we only select for
+     * RRScreenChangeNotifyMask.  We used to select for other values in
+     * RR*NotifyMask, but we weren't really doing anything useful with those
+     * events.  We only care about "the screens changed in some way or another"
+     * for now.
+     *
+     * If we ever run into a situtation that could benefit from processing more
+     * detailed events, we can enable this code again.
+     *
+     * Note that the X server sends RRScreenChangeNotify in conjunction with the
+     * more detailed events from RANDR 1.2 - see xserver/randr/randr.c:TellChanged().
+     */
     else if (event_num == RRNotify)
     {
 	/* Other RandR events */
@@ -522,6 +579,7 @@ screen_on_event (GdkXEvent *xevent,
 	/* No need to reprobe hardware here */
 	screen_update (screen, TRUE, FALSE, NULL); /* NULL-GError */
     }
+#endif
 
     /* Pass the event on to GTK+ */
     return GDK_FILTER_CONTINUE;
@@ -568,18 +626,18 @@ gnome_rr_screen_new (GdkScreen *gdk_screen,
 	    return NULL;
 	}
 
-	XRRSelectInput (screen->xdisplay,
-			screen->xroot,
-			RRScreenChangeNotifyMask	|
-			RRCrtcChangeNotifyMask		|
-			RROutputPropertyNotifyMask);
-	
-	gdk_x11_register_standard_event_type (
-	    gdk_screen_get_display (gdk_screen),
-	    event_base,
-	    RRNotify + 1);
-	
-	gdk_window_add_filter (screen->gdk_root, screen_on_event, screen);
+	if (screen->callback) {
+	    XRRSelectInput (screen->xdisplay,
+			    screen->xroot,
+			    RRScreenChangeNotifyMask);
+
+	    gdk_x11_register_standard_event_type (gdk_screen_get_display (gdk_screen),
+						  event_base,
+						  RRNotify + 1);
+
+	    gdk_window_add_filter (screen->gdk_root, screen_on_event, screen);
+	}
+
 	return screen;
     }
     else
@@ -639,6 +697,33 @@ gnome_rr_screen_get_ranges (GnomeRRScreen *screen,
 	*max_height = screen->info->max_height;
 }
 
+/**
+ * gnome_rr_screen_get_timestamps
+ * @screen: a #GnomeRRScreen
+ * @change_timestamp_ret: Location in which to store the timestamp at which the RANDR configuration was last changed
+ * @config_timestamp_ret: Location in which to store the timestamp at which the RANDR configuration was last obtained
+ *
+ * Queries the two timestamps that the X RANDR extension maintains.  The X
+ * server will prevent change requests for stale configurations, those whose
+ * timestamp is not equal to that of the latest request for configuration.  The
+ * X server will also prevent change requests that have an older timestamp to
+ * the latest change request.
+ */
+void
+gnome_rr_screen_get_timestamps (GnomeRRScreen *screen,
+				guint32       *change_timestamp_ret,
+				guint32       *config_timestamp_ret)
+{
+    g_return_if_fail (screen != NULL);
+
+    if (change_timestamp_ret)
+	*change_timestamp_ret = screen->info->resources->timestamp;
+
+    if (config_timestamp_ret)
+	*config_timestamp_ret = screen->info->resources->configTimestamp;
+}
+
+
 /**
  * gnome_rr_screen_refresh
  * @screen: a #GnomeRRScreen
@@ -1092,6 +1177,20 @@ gnome_rr_crtc_set_config (GnomeRRCrtc      *crtc,
 			  GnomeRROutput   **outputs,
 			  int               n_outputs,
 			  GError          **error)
+{
+    return gnome_rr_crtc_set_config_with_time (crtc, GDK_CURRENT_TIME, x, y, mode, rotation, outputs, n_outputs, error);
+}
+
+gboolean
+gnome_rr_crtc_set_config_with_time (GnomeRRCrtc      *crtc,
+				    guint32           timestamp,
+				    int               x,
+				    int               y,
+				    GnomeRRMode      *mode,
+				    GnomeRRRotation   rotation,
+				    GnomeRROutput   **outputs,
+				    int               n_outputs,
+				    GError          **error)
 {
     ScreenInfo *info;
     GArray *output_ids;
@@ -1133,7 +1232,7 @@ gnome_rr_crtc_set_config (GnomeRRCrtc      *crtc,
     }
     
     status = XRRSetCrtcConfig (DISPLAY (crtc), info->resources, crtc->id,
-			       CurrentTime, 
+			       timestamp, 
 			       x, y,
 			       mode ? mode->id : None,
 			       xrotation_from_rotation (rotation),
