@@ -32,6 +32,7 @@
 #include <gdk/gdk.h>
 #include <gconf/gconf-client.h>
 #include <libgnomeui/gnome-desktop-thumbnail.h>
+#include <libgnomeui/gnome-bg-crossfade.h>
 
 G_BEGIN_DECLS
 
@@ -105,10 +106,16 @@ gboolean         gnome_bg_is_dark               (GnomeBG               *bg);
 gboolean         gnome_bg_changes_with_size     (GnomeBG               *bg);
 
 
-/* Set a pixmap as root - not a GnomeBG method */
+/* Set a pixmap as root - not a GnomeBG method. At some point
+ * if we decide to stabilize the API then we may want to make
+ * these object methods, drop gnome_bg_create_pixmap, etc.
+ */
 void             gnome_bg_set_pixmap_as_root    (GdkScreen             *screen,
 						 GdkPixmap             *pixmap);
 
+GnomeBGCrossfade *gnome_bg_set_pixmap_as_root_with_crossfade (GdkScreen *screen,
+                                                              GdkPixmap *pixmap);
+GdkPixmap *gnome_bg_get_pixmap_from_root (GdkScreen *screen);
 
 G_END_DECLS
 
